// FlowMoBlur - use flow to artifically motion blur
//
// java FlowMoBlur Flowpaintele/comp.1-7.ppm flow_ 2.5

import java.io.*;
import java.util.*;
import java.text.*;

import IF.FrameSpec;
import GR.*;
import zlib.*;
import gnu.*;

public class FlowMoBlur
{
  public static void main(String[] cmdline)
  {
    try {
      Args args = new Args(cmdline);
      args.parse("+:str:plateseq:"+
		 "+:str:flowprefix, <frame>.nnnn.flo is appended:"+
		 "+:flt:scale motion blur vector in pixels:"+
		 "-flipflow:flg:flip flowfield direction:");
      //"-flowflip:flg:flip flowfield direction";

      String platespec = args.getRequiredString(0);
      FrameSpec ps = new FrameSpec(platespec);

      String flowprefix = args.getRequiredString(1);

      float scale = 1.f;
      Float Scale = args.getRequiredFloat(2);
      if (Scale != null) {
	scale = Scale.floatValue();
	System.out.println("scaling by " + scale);
      }
      else
	System.exit(1);

      boolean flipflow = false;
      Boolean oflipflow = args.getFlag("-flowflip");
      if (oflipflow != null) flipflow = oflipflow.booleanValue();

      int f1 = ps.getFirstFrame();
      int fl = ps.getLastFrame();

      DecimalFormat df = new DecimalFormat("0000");

      for( int f = f1; f <= fl; f++ ) {

	// read next frame in sequence,
	String platepath = ps.getFullPathOfFrame(f);
	gr[] curframe = grUtil.loadByExt(platepath);
	gr[] blurframe = grUtil.makecongruent("blurframe", curframe);

	//String flowframe = flowprefix + f1;
	float[][][] uvc
	  = FlowUtil.loadFlow(flowprefix +"."+ df.format(f) + ".flo");
	grFloatArr uc = new grFloatArr("uc", uvc[0]);
	grFloatArr vc = new grFloatArr("vc", uvc[1]);
	//grFloatArr uc = grfArrUtil.load(flowframe + "_u.flt");
	//grFloatArr vc = grfArrUtil.load(flowframe + "_v.flt");

	moblur(curframe, uc.getData(), vc.getData(), scale, flipflow,
	       blurframe);

	String fs = df.format(f);
	System.out.println("saving frame " + f);
	grUtil.saveByExt(blurframe, "comp."+fs+".ppm");

      } //loop

      System.exit(0);
    }
    catch(Exception x) { zliberror.die(x); }
  }

  // 
  static void moblur(gr[] src, float[][] uf, float[][] vf, float scale,
		     boolean flipflow,
		     gr[] dst)
  {
    int yres = uf.length;
    int xres = uf[0].length;
    if (src[0].xres() < xres) xres = src[0].xres();
    if (src[0].yres() < yres) yres = src[0].yres();
    zliberror.assert(dst[0].xres() >= xres);
    zliberror.assert(dst[0].yres() >= yres);

    final boolean trace = false;
    final boolean trace2 = false;
    for( int y=0; y < yres; y++ ) {
      for( int x=0; x < xres; x++ ) {

	float u = uf[y][x];
	float v = vf[y][x];

	if (flipflow) {
	  u = -u;
	  v = -v;
	}

	u = u * scale;
	v = v * scale;
	float len = (float)Math.sqrt(u*u + v*v);

//  	if (trace == false && (len/scale)>2.f) {
//  	  trace = true;
//  	  trace2 = true;
//  	} else trace2 = false;

	// integrate vector of length len with sampsPerPix samples per pixel
	float sampsPerPix = 1.5f;
	int ilen = (int)Math.ceil(sampsPerPix*len);

	//if (trace2)
	//System.out.println("scale="+scale+" len="+len+" ilen="+ilen+" u="+u);
	
	float rsum = 0.f, gsum = 0.f, bsum = 0.f;
	float ilenm1 = ilen - 1;
	for( int i = 0; i < ilen; i++ ) {
	  float frac = 0.f;
	  if (ilen > 1) frac = (float)i / ilenm1;
	  float xu = x + frac*u;
	  float yv = y + frac*v;
	  //if (trace2) System.out.println(" xu "+xu);

	  rsum += grUtil.bilerp(src[0], xu, yv);
	  if (src.length >= 3) {
	    gsum += grUtil.bilerp(src[1], xu, yv);
	    bsum += grUtil.bilerp(src[2], xu, yv);
	  }
	} //i along motion vector

	rsum /= (float)ilen;
	gsum /= (float)ilen;
	bsum /= (float)ilen;

	dst[0].wr(x, y, (int)rsum);
	if (dst.length >= 3) {
	  dst[1].wr(x, y, (int)gsum);
	  dst[2].wr(x, y, (int)bsum);
	}

      } //x
    } //y

  } //fade

} //FlowMoBlur
