// FlowSloMo - use flow to make inbetween frames.
// we just move curframe forwared by the lerped flow
// rather than crossfading.  This may be better - no double images.
//
// java FlowSloMo Flowpaintele/comp.1-7.ppm flow_ 5

import java.io.*;
import java.util.*;
import java.text.*;

import IF.FrameSpec;
import GR.*;
import zlib.*;
import gnu.*;

public class FlowSloMo
{
  static int	_verbose = 1;

  public static void main(String[] cmdline)
  {
    try {
      Args args = new Args(cmdline);
      args.parse("+:str:plateseq:"+
		 "+:str:flowprefix, <frame>.flo is appended:"+
		 "+:int:lerp flow to slow mo.  1=full speed, 2=half, etc:");
      //"-flowflip:flg:flip flowfield direction";

      String platespec = args.getRequiredString(0);
      FrameSpec ps = new FrameSpec(platespec);

      String flowprefix = args.getRequiredString(1);

      int slow = 1;
      Integer Slow = args.getRequiredInteger(2);
      if (Slow != null) {
	slow = Slow.intValue();
	System.out.println("slowing down by " + slow);
      }
      else
	System.exit(1);

      int f1 = ps.getFirstFrame();
      int fl = ps.getLastFrame();

      DecimalFormat df = new DecimalFormat("0000");

      // read next frame in sequence,
      String platepath = ps.getFullPathOfFrame(f1);
      gr[] curframe = grUtil.loadByExt(platepath);
      gr[] framewarp = grUtil.makecongruent("framewarp", curframe);

      // load first flow in order to lazily make second arrays u,v
      float[][][] uvc
	= FlowUtil.loadFlow(flowprefix +"."+ df.format(f1) + ".flo");
      grFloatArr uc = new grFloatArr("uc", uvc[0]);
      grFloatArr vc = new grFloatArr("vc", uvc[1]);
      //grFloatArr uc = grfArrUtil.load(flowframe + "_u.flt");
      //grFloatArr vc = grfArrUtil.load(flowframe + "_v.flt");

      grFloatArr u = grfArrUtil.makeCongruent("u", uc);
      grFloatArr v = grfArrUtil.makeCongruent("v", vc);

      int of = f1;
      for( int f = f1; f <= fl; f++ ) {

	// assume flowfield at frame f means flow from f to f+1
	uvc = FlowUtil.loadFlow(flowprefix +"."+ df.format(f) + ".flo");
	uc = new grFloatArr("uc", uvc[0]);
	vc = new grFloatArr("vc", uvc[1]);
	//uc = grfArrUtil.load(flowframe + "_u.flt");
	//vc = grfArrUtil.load(flowframe + "_v.flt");

	// read next frame in sequence
	// TODO: 'nxtframe' not used currently
	int fn =  (f < fl) ? (f+1) : f;
	platepath = ps.getFullPathOfFrame(fn);
	gr[] nxtframe = grUtil.loadByExt(platepath);

	for( int si = 0; si < slow; si++ ) {

	  float frac = (float)si / (float)slow;
	  fade(uc.getData(), frac, u.getData());
	  fade(vc.getData(), frac, v.getData());

	  // we just move curframe forwared by the lerped flow
	  // rather than crossfading.  This may be better - no double images.
	  grOps.warp(curframe, framewarp, u.getData(), v.getData());

	  String fs = df.format(of);
	  System.out.println("saving frame " + of);
	  grUtil.saveByExt(framewarp, "comp."+fs+".ppm");

	  of++;
	}

	curframe = nxtframe;
      } //loop

      System.exit(0);
    }
    catch(Exception x) { zliberror.die(x); }
  }

  // fade the flow field if doing slowmo
  static void fade(float[][] a, float frac, float[][] o)
  {
    int yres = a.length;
    int xres = a[0].length;

    System.out.println("flowslomo direction seems right, need to verify?");
    for( int y=0; y < yres; y++ ) {
      for( int x=0; x < xres; x++ ) {
	float fa = a[y][x];		// direction- seems right.
	o[y][x] = frac * fa;
      }
    }

    if (_verbose > 0) {
      System.out.println("fade = "+frac);
      grfArrUtil.stats("original flow stats", a, null);
      grfArrUtil.stats("faded flow stats", o, null);
    }
  } //fade

} //FlowSloMo
